#!/usr/bin/env python3
###############################################################################
#
# The MIT License (MIT)
#
# Copyright (c) Tavendo GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
###############################################################################

# from pprint import pprint
from bts.ws.statistics_protocol import StatisticsProtocol
from bts.http_rpc import HTTPRPC

try:
    import asyncio
except ImportError:
    import trollius as asyncio


def id_to_int(id):
    return int(id.split('.')[-1])


class TradeProtocol(StatisticsProtocol):
    asset_info = {}

    def get_asset_info(self, asset_id):
        if asset_id not in self.asset_info:
            _asset_info = self.node_api.get_objects([asset_id])[0]
            self.asset_info[asset_id] = _asset_info
            self.asset_info[_asset_info["symbol"]] = _asset_info
        return self.asset_info[asset_id]

    def onTrade(self, trx):
        print("sent %s" % trx)

    def process_operations(self, op_id):
        op_info = self.node_api.get_objects([op_id])
        for operation in op_info[::-1]:
            if operation["op"][0] != 4:
                return
            op = operation["op"][1]
            trx = {}

            trx["block_num"] = operation["block_num"]
            block_info = self.node_api.get_block(trx["block_num"])
            trx["timestamp"] = block_info["timestamp"]
            trx["trx_id"] = operation["id"]
            # Get trade info
            for _type in ["pays", "receives", "fee"]:
                trx[_type] = [0, ""]
                asset_info = self.get_asset_info(op[_type]["asset_id"])
                trx[_type][1] = asset_info["symbol"]
                trx[_type][0] = float(op[_type]["amount"])/float(
                        10**int(asset_info["precision"]))

            self.onTrade(trx)


if __name__ == '__main__':
    import sys
    uri = ""
    if len(sys.argv) >= 2:
        uri = sys.argv[1]

    ws = TradeProtocol(uri)
    node_api = HTTPRPC(uri)
    ws.init_statistics(
        node_api, "exchange.btsbots")
    asyncio.get_event_loop().run_until_complete(ws.handler())



# -*- coding: utf-8 -*-
# from pytest import raises

# The parametrize function is generated, so this doesn't work:
#
#     from pytest.mark import parametrize
#
import pytest
parametrize = pytest.mark.parametrize
from bts import HTTPRPC
from pprint import pprint


class TestMain(object):
    logfile = open("/tmp/test-python-bts.log", 'a')
    host = "localhost"
    node_port = "4090"
    cli_port = "4092"
    cli_rpc = HTTPRPC(host, cli_port, "", "")

    def test_cli_rpc(self):
        pprint("======= test_cli_rpc =========", self.logfile)
        result = self.cli_rpc.about()
        pprint(result, self.logfile)
        result = self.cli_rpc.get_dynamic_global_properties()
        pprint(result, self.logfile)
        assert result["head_block_number"] > 0



#!/usr/bin/env python3
###############################################################################
#
# The MIT License (MIT)
#
# Copyright (c) Tavendo GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
###############################################################################

# from pprint import pprint
from bts.ws.base_protocol import BaseProtocol
from bts.http_rpc import HTTPRPC

try:
    import asyncio
except ImportError:
    import trollius as asyncio


def id_to_int(id):
    return int(id.split('.')[-1])


class StatisticsProtocol(BaseProtocol):
    account = {"name": "exchange.btsbots", "id": "", "statistics": ""}
    last_trx = ""
    last_op = "2.9.1"
    node_api = None

    def init_statistics(self, node_api, account_name):
        self.node_api = node_api
        self.account["name"] = account_name

    def process_operations(self, op_id):
        op_info = self.node_api.get_objects([op_id])
        print(op_info)

    def onStatistics(self, notify):
        # TODO: if network is ont sync, return
        trx_last = self.last_trx
        trx_current = notify["most_recent_op"]
        if id_to_int(trx_current) > id_to_int(trx_last):
            self.last_trx = trx_current
        else:
            return
        while True:
            if id_to_int(trx_current) <= id_to_int(trx_last):
                return
            trx_info = self.node_api.get_objects([trx_current])[0]
            if id_to_int(trx_info["operation_id"]) <= id_to_int(self.last_op):
                return
            self.process_operations(trx_info["operation_id"])
            trx_current = trx_info["next"]

    @asyncio.coroutine
    def onOpen(self):
        yield from super().onOpen()
        response = yield from self.rpc(
            [self.database_api, "get_account_by_name", [self.account["name"]]])
        self.account["statistics"] = response["statistics"]
        self.account["id"] = response["id"]
        statistics_info = self.node_api.get_objects(
            [self.account["statistics"]])[0]
        if self.last_trx == "":
            self.last_trx = statistics_info["most_recent_op"]
        print("monitor account %s, begin from trx: %s" % (
            self.account["name"], self.last_trx))
        self.onStatistics(statistics_info)
        self.subscribe(self.account["statistics"], self.onStatistics)


if __name__ == '__main__':
    import sys
    uri = ""
    if len(sys.argv) >= 2:
        uri = sys.argv[1]

    ws = StatisticsProtocol(uri)
    node_api = HTTPRPC(uri)
    ws.init_statistics(
        node_api, "exchange.btsbots")
    asyncio.get_event_loop().run_until_complete(ws.handler())



#!/usr/bin/env python3
###############################################################################
#
# The MIT License (MIT)
#
# Copyright (c) Tavendo GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
###############################################################################

# from pprint import pprint
import json
import websockets

try:
    import asyncio
except ImportError:
    import trollius as asyncio


class RPCError(Exception):
    pass


class BaseProtocol(object):
    def __init__(self, uri=""):
        if not uri:
            uri = "wss://bitshares.openledger.info/ws"
        self.uri = uri
        self.request_id = 0
        self.history_api = 0
        self.network_api = 0
        self.database_api = 0
        self.result = {}
        self.callbacks = {}

    async def rpc(self, params):
        request_id = self.request_id
        self.request_id += 1
        request = {"id": request_id, "method": "call", "params": params}
        future = self.result[request_id] = asyncio.Future()
        await self.websocket.send(json.dumps(request).encode('utf8'))
        await asyncio.wait_for(future, None)
        self.result.pop(request_id)
        ret = future.result()
        if 'error' in ret:
            if 'detail' in ret['error']:
                raise RPCError(ret['error']['detail'])
            else:
                raise RPCError(ret['error']['message'])
        return ret["result"]

    def subscribe(self, object_id, callback):
        if object_id not in self.callbacks:
            self.callbacks[object_id] = [callback]
        else:
            self.callbacks[object_id].append(callback)

    async def handler_message(self):
        while True:
            payload = await self.websocket.recv()
            self.onMessage(payload)

    async def onOpen(self):
        await self.rpc([1, "login", ["", ""]])
        self.database_api = await self.rpc([1, "database", []])
        self.history_api = await self.rpc([1, "history", []])
        self.network_api = await self.rpc([1, "network_broadcast", []])
        await self.rpc(
            [self.database_api, "set_subscribe_callback", [200, False]])

    async def handler(self):
        # can handle message less than 8M
        async with websockets.connect(
                self.uri, max_size=2**20*8, max_queue=2**5*2) as websocket:
            print("WebSocket connection open.")
            self.websocket = websocket
            task1 = asyncio.ensure_future(self.handler_message())
            task2 = asyncio.ensure_future(self.onOpen())
            await asyncio.wait([task1, task2])

    def onMessage(self, payload):
        res = json.loads(payload)
        if "id" in res and res["id"] in self.result:
            self.result[res["id"]].set_result(res)
        elif "method" in res:
            for notice in res["params"][1][0]:
                if "id" not in notice:
                    # means the object have removed from chain
                    if "removed" in self.callbacks:
                        for _cb in self.callbacks["removed"]:
                            _cb(notice)
                    continue
                for _id in self.callbacks:
                    if _id == notice["id"][:len(_id)]:
                        for _cb in self.callbacks[_id]:
                            _cb(notice)


if __name__ == '__main__':
    import sys
    uri = ""
    if len(sys.argv) >= 2:
        uri = sys.argv[1]

    ws = BaseProtocol(uri)
    asyncio.get_event_loop().run_until_complete(ws.handler())



# -*- coding: utf-8 -*-

from bts.ws.base_protocol import BaseProtocol  # noqa
from bts.ws.statistics_protocol import StatisticsProtocol  # noqa
from bts.ws.transfer_protocol import TransferProtocol  # noqa
